/*
    YAM - Yet Another Mod player
    Copyright 2001, 2002 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdlib.h>
#include <time.h>
#include "modplayer.h"
#include "convert.h"

ModPlayer::ModPlayer(ModFile *mf, int ch, int sr, int bd, bool s, bool le) {
	int i, n;

	srand(time(NULL));

	mod = mf;
	outChannels = ch;
	samplerate = sr;
	bitdepth = bd;
	sig = s;
	littleendian = le;

	orderNr = currentPattern = patternOffset = 0;
	loopRow = loopNum = patternDelay = 0;

	patternCallback = NULL;
	rowCallback = NULL;
	hitCallback = NULL;

	lastInstrument = (int*) malloc(sizeof(int)*mod->channels);
	origVolume = (int*) malloc(sizeof(int)*mod->channels);
	volume = (int*) malloc(sizeof(int)*mod->channels);
	tone = (int*) malloc(sizeof(int)*mod->channels);
	amiga = (int*) malloc(sizeof(int)*mod->channels);
	frequency = (int*) malloc(sizeof(int)*mod->channels);
	pan = (int*) malloc(sizeof(int)*mod->channels);
	lastPos = (int*) malloc(sizeof(int)*mod->channels);

	portaSpeed = (int*) malloc(sizeof(int)*mod->channels);
	portaNote = (int*) malloc(sizeof(int)*mod->channels);
	vibratoPos = (int*) malloc(sizeof(int)*mod->channels);
	vibratoNeg = (bool*) malloc(sizeof(bool)*mod->channels);
	vibratoPar = (int*) malloc(sizeof(int)*mod->channels);
	vibratoType = (int*) malloc(sizeof(int)*mod->channels);
	tremoloPos = (int*) malloc(sizeof(int)*mod->channels);
	tremoloNeg = (bool*) malloc(sizeof(bool)*mod->channels);
	tremoloPar = (int*) malloc(sizeof(int)*mod->channels);
	tremoloType = (int*) malloc(sizeof(int)*mod->channels);
	glissando = (bool*) malloc(sizeof(bool)*mod->channels);
	loopRow = 0;
	loopNum = 0;
	patternDelay = 0;

	for (i = 0; i < mod->channels; i++) {
		n = i%4;
		if (n==0 || n==3) pan[i] = 64;
		else pan[i] = 192;

		lastInstrument[i] = -1;

		origVolume[i] = volume[i] = 0;
		tone[i] = amiga[i] = frequency[i] = 0;
		portaSpeed[i] = portaNote[i] = 0;
		vibratoPos[i] = vibratoPar[i] = vibratoType[i] = 0;
		tremoloPos[i] = tremoloPar[i] = tremoloType[i] = 0;
		vibratoNeg[i] = tremoloNeg[i] = glissando[i] = false;

		lastPos[i] = 0;
	}

	running = true;
	started = true;
	fade = false;
	fadePos = 0;

	speed = 6;
	bpm = 125;
	updateSpeed();

	tick = speed;
	row = 64;
	orderNr = -1;
	tickPos = tickLength;
}

ModPlayer::~ModPlayer(void) {
//	delete mod;
	free(lastInstrument);
	free(origVolume);
	free(volume);
	free(tone);
	free(amiga);
	free(frequency);
	free(pan);
	free(portaSpeed);
	free(portaNote);
	free(vibratoPos);
	free(vibratoPar);
	free(vibratoType);
	free(tremoloPos);
	free(tremoloPar);
	free(tremoloType);
	free(vibratoNeg);
	free(tremoloNeg);
	free(glissando);
	free(lastPos);
}

void ModPlayer::updateSpeed(void) {
	int hz;
//	int length;

	hz = 2*bpm/5;
//	length = 1000/hz;
	frameSize = getFrameSize();

	tickLength = samplerate/hz;
}

int ModPlayer::read(unsigned char *buffer, int length) {
	int numSamples;
	int i, j;
	int l, r;
	int sampleNum, samplePos;
	sample *sPtr;
	int data;

	numSamples = length/frameSize;

	for (i = 0; i < numSamples; i++) {
		if (tickPos>=tickLength) {
			tickPos = 0;
			for (j = 0; j < mod->channels; j++)
				lastPos[j] += tickLength*frequency[j]/samplerate;
			timer();
		}

		l = r = 0;

		if (running) {
			for (j = 0; j < mod->channels; j++) {

				sampleNum = lastInstrument[j];
				samplePos = lastPos[j] + tickPos*frequency[j]/samplerate;

				if (sampleNum<0) continue;
				sPtr = &(mod->samples[sampleNum]);

				if ((sPtr->loopLength>0) && (samplePos >= sPtr->loopStart + sPtr->loopLength)) {
					samplePos -= sPtr->loopLength;
					lastPos[j] -= sPtr->loopLength;
				}

				if (samplePos >= sPtr->length) continue;

				// <<1
				data = (sPtr->data[samplePos]*volume[j]);

				if (pan[j]<127) {
					l += data;
					r += pan[j]*data/127;
				} else {
					l += (255-pan[j])*data/127;
					r += data;
				}

//				l += (255-pan[j])*data/255;
//				r += pan[j]*data/255;

			}
		} else {
			break;
		}

		if (fade) {
			if (fadePos>=fadeLength) {
				started = false;
				running = false;
				break;
			} else {
				l = l*(fadeLength-fadePos)/fadeLength;
				r = r*(fadeLength-fadePos)/fadeLength;
				fadePos++;
			}
		}

		convert(&buffer,l,r);

		tickPos++;

	}

	return i*frameSize;
}

void ModPlayer::timer(void) {
	tick++;

	if (tick>=speed) {
		if (patternDelay==0) {
			updateRow();
			row++;
		} else patternDelay--;
		tick = 0;
	} else {
		updateEffect();
	}
}

void ModPlayer::updateRow(void) {
	bool jumpToNext, jumpToPattern;
	int nextRow, nextPatternNr;
	int i;
	note *n;
	int subtype, parameter;

	if (row>=64) nextPattern();

	if (!running) return;

	if (rowCallback!=NULL) rowCallback(row,currentPattern,orderNr,mod->songLength);

	patternOffset = currentPattern*mod->channels*64 + row*mod->channels;

	jumpToNext = false;
	nextRow = 0;
	jumpToPattern = false;
	nextPatternNr = 0;

	for (i = 0; i < mod->channels; i++) {
		n = &(mod->patterns[patternOffset+i]);

		if (n->sample>0 && n->sample<=mod->numSamples) {
			lastInstrument[i] = n->sample-1;
			origVolume[i] = mod->samples[lastInstrument[i]].volume;
		}

		if (n->tone>0 && n->effect != 0x3 && n->effect != 0x5 && (n->effect != 0xE || ((n->parameter>>4)!=0xD))) {
			lastPos[i] = 0;
			tone[i] = n->tone + mod->samples[lastInstrument[i]].finetune;
			amiga[i] = toneToAmiga(tone[i]);
			frequency[i] = toFreq(amiga[i]);
			if (vibratoType[i]>=4) {
				vibratoPos[i] = 0;
				vibratoNeg[i] = false;
			}
			if (tremoloType[i]>=4) {
				tremoloPos[i] = 0;
				tremoloNeg[i] = false;
			}
			if (hitCallback!=NULL) hitCallback(i);
		}

		if (n->effect!=0 || n->parameter!=0) {

			switch (n->effect) {
				case 0xD:
					// Pattern Break
					if (n->parameter < 0x40 && n->parameter >= 0)
						nextRow = n->parameter;
					else
						nextRow = 0;
					jumpToNext = true;
					break;
				case 0xF:
					// Set speed
					if (n->parameter <= 0x1F && n->parameter >= 0x00) {
						speed = n->parameter;
					} else if (n->parameter >= 0x20 && n->parameter <= 0xFF) {
						bpm = n->parameter;
						updateSpeed();
					}
					break;
				case 0xC:
					// Set volume
					if (n->parameter < 0)
						origVolume[i] = 0;
					else if (n->parameter > 0x40)
						origVolume[i] = 0x40;
					else
						origVolume[i] = n->parameter;
					break;
				case 0xB:
					// Jump To Pattern
					if (n->parameter >= 0 && n->parameter < mod->songLength) {
//						jumpToPattern = true;
						nextPatternNr = n->parameter;
					}
					break;
				case 0x3:
					// Porta To Note
					if (n->tone>0)
						portaNote[i] = toneToAmiga(n->tone + mod->samples[lastInstrument[i]].finetune);
					if (n->parameter>0)
						portaSpeed[i] = n->parameter;
					break;
				case 0x4:
					// Vibrato
					if (n->parameter>0)
						vibratoPar[i] = n->parameter;
					break;
				case 0x8:
					// Pan
					if (n->parameter<=0x80) {
						pan[i] = 255*n->parameter/0x80;
					} else if (n->parameter==0xA4) {
					}
					break;
				case 0x7:
					// Tremolo
					if (n->parameter>0)
						tremoloPar[i] = n->parameter;
					break;
				case 0x9:
					// Sample offset
					lastPos[i] += n->parameter<<8;
					if (lastPos[i]>=mod->samples[lastInstrument[i]].length) lastPos[i] -= n->parameter<<8;
					break;
				case 0xE:
					subtype = (n->parameter&0xF0)>>4;
					parameter = n->parameter&0x0F;
					switch (subtype) {
						case 0x0:
							// Set filter
							break;
						case 0x1:
							// Fine Porta Up
							amiga[i] -= parameter;
							break;
						case 0x2:
							// Fine Porta Down
							amiga[i] += parameter;
							break;
						case 0x3:
							// Glissando Control
							if (parameter==0x1) glissando[i] = true;
							else glissando[i] = false;
							break;
						case 0x4:
							// Vibrato Waveform
							vibratoType[i] = parameter;
							break;
						case 0x5:
							// Set Finetune
							if (parameter>7)
								mod->samples[lastInstrument[i]].finetune = (char) (parameter-16);
							else
								mod->samples[lastInstrument[i]].finetune = (char) parameter;
							break;
						case 0x6:
							// Pattern Loop
							if (parameter==0) loopRow = row;
							else {
								if (loopNum==0) loopNum = parameter;
								else loopNum--;
								if (loopNum>0) row = loopRow-1;
							}
							break;
						case 0x7:
							// Tremolo Waveform
							tremoloType[i] = parameter;
							break;
						case 0x8:
							// 16 Pos Panning
							pan[i] = 255*parameter/15;
							break;
						case 0xA:
							// Fine VolSlide Up
							origVolume[i] += parameter;
							if (origVolume[i]>0x40) origVolume[i] = 0x40;
							break;
						case 0xB:
							// Fine VolSlide Down
							origVolume[i] -= parameter;
							if (origVolume[i]<0) origVolume[i] = 0;
							break;
						case 0xE:
							// Pattern Delay
							patternDelay = parameter;
							break;
						case 0xF:
							// Invert Loop
							break;
					}
					break;
			}


		}
		frequency[i] = toFreq(amiga[i]);
		volume[i] = origVolume[i];
	}

	if (jumpToNext) {
		nextPattern();
		row = nextRow-1;
	}
	if (jumpToPattern) {
		orderNr = nextPatternNr;
		currentPattern = mod->orders[orderNr];
		row = -1;
	}
}

void ModPlayer::nextPattern(void) {
	row = 0;
	orderNr++;
	if (orderNr>=mod->songLength) {
		running = false;
		started = false;
		return;
	}
	currentPattern = mod->orders[orderNr];
	if (patternCallback!=NULL) patternCallback(orderNr);
}

const static short sineTable[] = {
	  0, 24, 49, 74, 97,120,141,161,
	180,197,212,224,235,244,250,253,
	255,253,250,244,235,224,212,197,
	180,161,141,120, 97, 74, 49, 24,
};

void ModPlayer::updateEffect(void) {
	int i;
	note *n;
	int subtype, parameter;
	int depth, effectSpeed, delta;

	for (i = 0; i < mod->channels; i++) {
		n = &(mod->patterns[patternOffset+i]);

		if (n->effect!=0 || n->parameter!=0) {

			if (n->effect == 0xA || n->effect == 0x5 || n->effect == 0x6) {
				// Volume Slide
				if ((n->parameter&0xF0)==0 && (n->parameter&0x0F)>0)
					origVolume[i] -= n->parameter&0x0F;
				else if ((n->parameter&0xF0)>0 && (n->parameter&0xF0)==0)
					origVolume[i] += (n->parameter&0xF0)>>4;

				if (origVolume[i] > 0x40)
					origVolume[i] = 0x40;
				else if (origVolume[i] < 0)
					origVolume[i] = 0;

			}

			switch (n->effect) {
				case 0x1:
					// Porta Up
					amiga[i] -= n->parameter;
					// 54
					if (amiga[i]<5) amiga[i] = 5;
					break;
				case 0x2:
					// Porta Down
					amiga[i] += n->parameter;
					break;
				case 0x3:
				case 0x5:
					// Porta To Note
					if (amiga[i] > portaNote[i]) {
						if (!glissando[i]) {
							amiga[i] -= portaSpeed[i];
						} else {
							tone[i] += 8;
							amiga[i] = toneToAmiga(tone[i]);
						}
						if (amiga[i] < portaNote[i]) amiga[i] = portaNote[i];
					} else if (amiga[i] < portaNote[i]) {
						if (!glissando[i]) {
							amiga[i] += portaSpeed[i];
						} else {
							tone[i] -= 8;
							amiga[i] = toneToAmiga(tone[i]);
						}
						if (amiga[i] > portaNote[i]) amiga[i] = portaNote[i];
					}
					break;
				case 0x0:
					// Arpeggio
					switch (tick%3) {
						case 0:
							amiga[i] = toneToAmiga(tone[i]);
							break;
						case 1:
							amiga[i] = toneToAmiga(tone[i]+8*((n->parameter&0xF0)>>4));
							break;
						case 2:
							amiga[i] = toneToAmiga(tone[i]+8*(n->parameter&0x0F));
							break;
					}
					break;

				case 0xE:
					subtype = (n->parameter&0xF0)>>4;
					parameter = n->parameter&0x0F;
					switch (subtype) {
						case 0x9:
							// Retrig Note
							if (tick%parameter==0) lastPos[i] = 0;
							break;
						case 0xC:
							// Cut Note
							if (tick==parameter) origVolume[i] = 0;
							break;
						case 0xD:
							// Delay Note
							if (tick==parameter) {
								lastPos[i] = 0;
								tone[i] = n->tone + mod->samples[lastInstrument[i]].finetune;
								amiga[i] = toneToAmiga(tone[i]);
								frequency[i] = toFreq(amiga[i]);
								if (vibratoType[i]>=4) {
									vibratoPos[i] = 0;
									vibratoNeg[i] = false;
								}
								if (tremoloType[i]>=4) {
									tremoloPos[i] = 0;
									tremoloNeg[i] = false;
								}
								if (hitCallback!=NULL) hitCallback(i);
							}
							break;
					}
					break;
			}

		}

		volume[i] = origVolume[i];
		frequency[i] = toFreq(amiga[i]);

		// If using arpeggio, reset amiga-value, to get the real tone after the effect
		if (n->effect==0x0 && n->parameter!=0x00) amiga[i] = toneToAmiga(tone[i]);


		if (n->effect==0x4 || n->effect==0x6) {
			// Vibrato
			depth = vibratoPar[i]&0x0F;
			effectSpeed = (vibratoPar[i]&0xF0)>>4;
			delta = 0;
			switch (vibratoType[i]) {
				case 0:
				case 4:
					// Sine wave
					delta = sineTable[vibratoPos[i]];
					break;
				case 1:
				case 5:
					// Ramp down
					if (!vibratoNeg[i]) delta = (-8*vibratoPos[i]);
					else delta = -(255 - (8*vibratoPos[i]));
					break;
				case 2:
				case 6:
					// Square wave
					delta = 255;
					break;
				case 3:
				case 7:
					// Random
					delta = 255-(rand()%511);
					break;
			}

			delta = (depth*delta)>>7;

			if (!vibratoNeg[i]) frequency[i] = toFreq(amiga[i]+delta);
			else frequency[i] = toFreq(amiga[i]-delta);

			vibratoPos[i] += effectSpeed;
			if (vibratoPos[i]>=32) {
				vibratoPos[i] -= 32;
				vibratoNeg[i] = (vibratoNeg[i]) ? false : true;
			}
		} else if (n->effect==0x7) {
			// Tremolo
			depth = tremoloPar[i]&0x0F;
			effectSpeed = (tremoloPar[i]&0xF0)>>4;
			delta = 0;
			switch (tremoloType[i]) {
				case 0:
				case 4:
					// Sine wave
					delta = sineTable[tremoloPos[i]];
					break;
				case 1:
				case 5:
					// Ramp down
					if (!tremoloNeg[i]) delta = (-8*tremoloPos[i]);
					else delta = -(255 - (8*tremoloPos[i]));
					break;
				case 2:
				case 6:
					// Square wave
					delta = 255;
					break;
				case 3:
				case 7:
					// Random
					delta = 255-(rand()%511);
					break;
			}

			delta = (depth*delta)>>6;

			if (!tremoloNeg[i]) {
				volume[i] = origVolume[i]+delta;
				if (volume[i]>0x40) volume[i] = 0x40;
			} else {
				volume[i] = origVolume[i]-delta;
				if (volume[i]<0) volume[i] = 0;
			}

			tremoloPos[i] += effectSpeed;
			if (tremoloPos[i]>=32) {
				tremoloPos[i] -= 32;
				tremoloNeg[i] = (tremoloNeg[i]) ? false : true;
			}
		}

	}

}

void ModPlayer::addPatternCallback(void (*callback)(int orderNr)) {
	patternCallback = callback;
}

void ModPlayer::addRowCallback(void (*callback)(int row, int patternNr, int orderNr, int length)) {
	rowCallback = callback;
}

void ModPlayer::addHitCallback(void (*callback)(int channel)) {
	hitCallback = callback;
}

void ModPlayer::fadeOut(void) {
	fadeOut(1000);
}

void ModPlayer::fadeOut(int length) {
	fade = true;
	fadePos = 0;
	fadeLength = samplerate*length/1000;
}

bool ModPlayer::isRunning(void) {
	return running;
}

int ModPlayer_read(void *ptr, unsigned char *buffer, int length) {
	return ((ModPlayer*)ptr)->read(buffer,length);
}


