/*
    YAM - Yet Another Mod player
    Copyright 2001, 2002 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#ifndef __MODPLAYER_H
#define __MODPLAYER_H

#include "modfile.h"

class ModPlayer {
	public:
		ModPlayer(ModFile *mf, int channels=2, int samplerate=44100, int bitdepth=16, bool sig=true, bool littleendian=true);
		~ModPlayer(void);
		int read(unsigned char *buffer, int length);
		void addPatternCallback(void (*callback)(int orderNr));
		void addRowCallback(void (*callback)(int row, int patternNr, int orderNr, int length));
		void addHitCallback(void (*callback)(int channel));
		void fadeOut(void);
		void fadeOut(int length);
		bool isRunning(void);
	private:
		ModFile *mod;
		int outChannels, samplerate, bitdepth;
		bool sig, littleendian;

		bool running, started;
		int speed, tick, row, bpm;
		int tickLength, tickPos;
		int orderNr, currentPattern, patternOffset;
		int *lastInstrument, *origVolume, *volume;
		int *tone, *amiga, *frequency, *pan;

		int *portaSpeed, *portaNote;
		int *vibratoPos, *vibratoPar, *vibratoType;
		int *tremoloPos, *tremoloPar, *tremoloType;
		bool *vibratoNeg, *tremoloNeg, *glissando;
		int loopRow, loopNum, patternDelay;

		int frameSize, *lastPos;

		bool fade;
		int fadePos, fadeLength;

		void (*patternCallback)(int orderNr);
		void (*rowCallback)(int row, int patternNr, int orderNr, int length);
		void (*hitCallback)(int channel);

		void updateSpeed(void);
		void timer(void);
		void updateRow(void);
		void updateEffect(void);
		void nextPattern(void);
};

int ModPlayer_read(void *ptr, unsigned char *buffer, int length);

#endif
