/*
    YAM - Yet Another Mod player
    Copyright 2001, 2002 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include "audio.h"

#define DEVICE "/dev/audio"

static int audio;
static int stop;
static int playing = 0;

int channels = 1;
int samplerate = 8000;
int bitdepth = 16;
int sig = 1;
int littleendian = 1;

int init_audio(void) {
	if ((audio = open(DEVICE,O_WRONLY)) < 0) {
		perror(DEVICE);
		return 0;
	}
	return 1;
}

void stop_audio(void) {
	close(audio);
}

const static unsigned char lut[] = {
	0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
	4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
	5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
	5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
	6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
	6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
	6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
	6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
};

static unsigned char toUlaw(short in) {
	unsigned char sign, exponent, mantissa;
	sign = (in>>8)&0x80;
	if (sign!=0) in = -in;
	exponent = lut[(in>>7)&0xFF];
	mantissa = (in>>(exponent+3))&0x0F;
	return ~(sign | (exponent<<4) | mantissa);
}

static void convert(unsigned char *out, unsigned char *in, int n) {
	short *inPtr = (short*) in;
	int i;
	for (i = 0; i < n; i++)
		*out++ = toUlaw(*inPtr++);
}

void start_play(read_func rf, void *ptr) {
	unsigned char buffer[512*2];
	unsigned char out[512];
	int n;
	stop = 0;
	playing = 1;
	while (!stop) {
		n = rf(ptr,buffer,sizeof(buffer));
		if (n<=0) break;
		n >>= 1;
		convert(out,buffer,n);
		write(audio,out,n);
	}
	playing = 0;
}

int is_playing(void) {
	return playing;
}

void break_play(void) {
	stop = 1;
}

