/*
    YAM - Yet Another Mod player
    Copyright 2001, 2002 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <windows.h>
#include "audio.h"

// -lwinmm

#define BUFFERSIZE 32768

int channels = 2;
int samplerate = 44100;
int bitdepth = 16;
int sig = 1;
int littleendian = 1;

static HWAVEOUT out;

static int done = 0;
static int done2 = 0;

static read_func read;
static void *data;
static int stop;

void CALLBACK callback(HWAVEOUT hwo, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2) {
	WAVEHDR *hdr;
	if (stop) {
		done = done2 = 1;
		return;
	}
	if (uMsg==WOM_DONE) {
		// done playing this buffer, currently playing the other buffer,
		// refill this buffer and queue it
		if (done2) done = 1; // done playing the last block
		hdr = (WAVEHDR*) dwParam1;
		hdr->dwBufferLength = read(data,(unsigned char*)hdr->lpData,BUFFERSIZE);
		if (hdr->dwBufferLength==0) done2 = 1; // finished buffering audio
		else waveOutWrite(hwo,hdr,sizeof(WAVEHDR));
	}
}

int init_audio(void) {
	WAVEFORMATEX format;

	format.wBitsPerSample = bitdepth;
	format.nChannels = channels;
	format.nBlockAlign = channels*bitdepth/8;
	format.nSamplesPerSec = samplerate;
	format.nAvgBytesPerSec = format.nBlockAlign*samplerate;
	format.wFormatTag = WAVE_FORMAT_PCM;
	format.cbSize = 0;

	if (waveOutOpen(&out,WAVE_MAPPER,&format,(DWORD)callback,0,CALLBACK_FUNCTION)!=MMSYSERR_NOERROR) {
		printf("Can't open audio device\n");
		return 0;
	}
	return 1;
}

void start_play(read_func rf, void *ptr) {
	unsigned char buffer[BUFFERSIZE];
	unsigned char buffer2[BUFFERSIZE];
	WAVEHDR header1, header2;

	read = rf;
	data = ptr;
	stop = 0;

	header1.lpData = (char*) buffer;
	header1.dwBufferLength = sizeof(buffer);
	header1.dwUser = 0;
	header1.dwFlags = 0;
	header1.dwLoops = 0;
	header2.lpData = (char*) buffer2;
	header2.dwBufferLength = sizeof(buffer2);
	header2.dwUser = 0;
	header2.dwFlags = 0;
	header2.dwLoops = 0;

	waveOutPrepareHeader(out,&header1,sizeof(header1));
	waveOutPrepareHeader(out,&header2,sizeof(header2));
	header1.dwBufferLength = read(data,(unsigned char*)header1.lpData,BUFFERSIZE);
	header2.dwBufferLength = read(data,(unsigned char*)header2.lpData,BUFFERSIZE);
	waveOutWrite(out,&header1,sizeof(header1));
	if (header2.dwBufferLength>0) {
		// queue the two blocks, use the callback to requeue them
		waveOutWrite(out,&header2,sizeof(header2));
		while (!done) Sleep(10);
	} else {
		// only playing one block, wait until it has finished
		while (!done2) Sleep(10);
	}

	waveOutUnprepareHeader(out,&header1,sizeof(header1));
	waveOutUnprepareHeader(out,&header2,sizeof(header2));
}

void stop_audio(void) {
	waveOutClose(out);
}

void break_play(void) {
	stop = 1;
}

